/* Copyright (C) 2017 by Bundesanstalt für Gewässerkunde
 * Software engineering by
 *  Björnsen Beratende Ingenieure GmbH
 *  Dr. Schumacher Ingenieurbüro für Wasser und Umwelt
 *
 * This file is Free Software under the GNU AGPL (>=v3)
 * and comes with ABSOLUTELY NO WARRANTY! Check out the
 * documentation coming with Dive4Elements River for details.
 */
package org.dive4elements.river.artifacts.sinfo.flood_duration;

import java.io.Serializable;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.collections.Predicate;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.dive4elements.artifacts.CallContext;
import org.dive4elements.river.artifacts.common.AbstractCalculationExportableResult;
import org.dive4elements.river.artifacts.common.AbstractExportContext;
import org.dive4elements.river.artifacts.common.ExportContextCSV;
import org.dive4elements.river.artifacts.common.ExportContextPDF;
import org.dive4elements.river.artifacts.common.GeneralResultType;
import org.dive4elements.river.artifacts.common.IResultType;
import org.dive4elements.river.artifacts.common.MetaAndTableJRDataSource;
import org.dive4elements.river.artifacts.common.ResultRow;
import org.dive4elements.river.artifacts.resources.Resources;
import org.dive4elements.river.artifacts.sinfo.common.SInfoResultType;
import org.dive4elements.river.artifacts.sinfo.util.RiverInfo;
import org.dive4elements.river.model.Attribute.AttributeKey;

import gnu.trove.TDoubleArrayList;

/**
 * Contains the result of a {@link FloodDurationCalculation}.
 *
 * @author Gernot Belger
 */
public final class FloodDurationCalculationResult extends AbstractCalculationExportableResult {

    private final static class InfrastructurePredicate implements Predicate {

        private final Infrastructure m_infrastructure;

        public InfrastructurePredicate(final Infrastructure infrastructure) {
            this.m_infrastructure = infrastructure;
        }

        @Override
        public boolean evaluate(final Object object) {
            final ResultRow row = (ResultRow) object;
            final Infrastructure test = new Infrastructure(row);
            return this.m_infrastructure.equals(test);
        }
    }

    private final static class HasInfrastructurePredicate implements Predicate {

        @Override
        public boolean evaluate(final Object object) {
            final ResultRow row = (ResultRow) object;

            return row.getValue(SInfoResultType.riverside) != null;
        }
    }

    public static final class Infrastructure implements Serializable {

        private static final long serialVersionUID = 1L;

        private final String m_group;

        private final String m_type;

        private final AttributeKey m_riverside;

        private static final String FACET_FLOOD_DURATION_DESCRIPTION = "sinfo_facet_flood_duration";

        private static final String FACET_ABSOLUTE_HEIGHT = "sinfo.flood_duration.absolute.height";

        public Infrastructure(final ResultRow row) {
            this.m_group = String.valueOf(row.getValue(SInfoResultType.infrastructuregroup));
            this.m_type = String.valueOf(row.getValue(SInfoResultType.infrastructuretype));
            final String riversideStr = String.valueOf(row.getValue(SInfoResultType.riverside));
            this.m_riverside = riversideStr.equals("null") ? AttributeKey.NONE : AttributeKey.valueOf(riversideStr);
        }

        public AttributeKey getRiverside() {
            return this.m_riverside;
        }

        @Override
        public int hashCode() {
            return new HashCodeBuilder() //
                    .append(this.m_group)//
                    .append(this.m_type)//
                    .append(this.m_riverside)//
                    .toHashCode();
        }

        @Override
        public boolean equals(final Object obj) {

            if (obj == null)
                return false;
            if (obj == this)
                return true;
            if (obj.getClass() != getClass())
                return false;

            final Infrastructure other = (Infrastructure) obj;
            return new EqualsBuilder() //
                    .append(this.m_group, other.m_group) //
                    .append(this.m_type, other.m_type) //
                    .append(this.m_riverside, other.m_riverside) //
                    .isEquals();
        }

        public String getFloodHeightLabel(final CallContext context) {
            return Resources.getMsg(context.getMeta(), FACET_ABSOLUTE_HEIGHT, FACET_ABSOLUTE_HEIGHT)
                    + " " + SInfoResultType.getInfrastructureLabel(context, this.m_group, this.m_type, this.m_riverside);
        }

        public String getFloodDurationLabel(final CallContext context) {
            return Resources.getMsg(context.getMeta(), FACET_FLOOD_DURATION_DESCRIPTION, FACET_FLOOD_DURATION_DESCRIPTION)
                    + " " + SInfoResultType.getInfrastructureLabel(context, this.m_group, this.m_type, this.m_riverside);
        }
    }

    private static final long serialVersionUID = 1L;

    private final boolean isUseWspl;

    private final String[] waterlevelLabels;

    private final int maxWaterlevelPdf = 3;

    private final Set<Infrastructure> m_infrastructures;

    public interface ValueGetter {
        double getValue(DurationWaterlevel waterlevel);
    }

    private enum ExportMode {
        pdf, csv
    }

    public FloodDurationCalculationResult(final String label, final String[] mainvalueLabels, final Collection<ResultRow> rows, final boolean isUseWspl,
            final Set<Infrastructure> infrastructures) {
        super(label, rows);
        this.waterlevelLabels = mainvalueLabels;
        this.isUseWspl = isUseWspl;
        this.m_infrastructures = infrastructures;
    }

    public Set<Infrastructure> getInfrastructureMap() {
        return this.m_infrastructures;
    }

    /**
     * Collection of the result rows containing only the rows describing an infrastructure
     */
    // FIXME: bad to override, instead make new method 'getInfrastructureRows' or similar?
    @Override
    public Collection<ResultRow> getRows() {

        final Collection<ResultRow> rows = super.getRows();

        final List<ResultRow> infrasOnlyRows = new ArrayList<>();
        for (final ResultRow row : rows) {
            if (row.getValue(SInfoResultType.infrastructuregroup) != null && row.getValue(SInfoResultType.infrastructuretype) != null)
                infrasOnlyRows.add(row);
        }
        return Collections.unmodifiableCollection(infrasOnlyRows);
    }

    /**
     * Collection of all result rows
     */
    public Collection<ResultRow> getAllRows() {

        return super.getRows();
    }

    @Override
    protected void writeCSVResultMetadata(final ExportContextCSV exportContextCSV) {
        /* nothing to do, as we never have several results, everything is written into the general header */
    }

    @Override
    protected String getJasperFile() {
        if (this.getWaterlevelCount() <= 1)
            return "/jasper/templates/sinfo.floodduration.jrxml";

        return "/jasper/templates/sinfo.floodduration2.jrxml";
    }

    protected String[] formatRow(final AbstractExportContext exportContextCSV, final ResultRow row, final ExportMode mode) {

        final Collection<String> lines = new ArrayList<>(20);

        lines.add(exportContextCSV.formatRowValue(row, GeneralResultType.station));
        lines.add(exportContextCSV.formatRowValue(row, SInfoResultType.riverside));
        lines.add(exportContextCSV.formatRowValue(row, SInfoResultType.floodDuration));

        lines.add(exportContextCSV.formatRowValue(row, SInfoResultType.floodDischarge));

        lines.add(exportContextCSV.formatRowValue(row, SInfoResultType.infrastructureHeight));
        lines.add(exportContextCSV.formatRowValue(row, SInfoResultType.infrastructuregroup));
        lines.add(exportContextCSV.formatRowValue(row, SInfoResultType.infrastructuretype));

        final List<DurationWaterlevel> waterlevelList = (List<DurationWaterlevel>) row.getValue(SInfoResultType.customMultiRowColWaterlevel);

        final int expectedSizetoGaugelabel = lines.size() + (waterlevelList.size() < 2 ? 4 : 12); // removing columns only works for fixed
        // indices

        final NumberFormat wFormatter = exportContextCSV.getFlowDepthFormatter();
        final NumberFormat qFormatter = exportContextCSV.getQFormatter();
        final NumberFormat durFormatter = exportContextCSV.getFloodDurationFormatter();

        for (int i = 0; i < waterlevelList.size(); i++) {

            if (i == this.maxWaterlevelPdf && mode == ExportMode.pdf)
                break;

            final DurationWaterlevel item = waterlevelList.get(i);
            lines.add(item.getWFormatted(wFormatter));
            lines.add(item.getFloodDurDaysPerYearFormatted(durFormatter));
            lines.add(item.getQFormatted(qFormatter));
            lines.add(item.getBezeichnung());
        }

        final int lineSize = lines.size();
        while (mode == ExportMode.pdf && lines.size() < expectedSizetoGaugelabel)
            lines.add("");

        if (isUseWspl() || mode == ExportMode.pdf)
            lines.add(exportContextCSV.formatRowValue(row, GeneralResultType.gaugeLabel)); // PDF: necessary, because if removed, location would not be shown!

        lines.add(exportContextCSV.formatRowValue(row, GeneralResultType.location));

        return lines.toArray(new String[lines.size()]);
    }

    @Override
    public void writeCSVHeader(final ExportContextCSV exportContextCSV, final RiverInfo river) {

        final Collection<String> header = new ArrayList<>(20);

        header.add(exportContextCSV.formatCsvHeader(GeneralResultType.station));
        header.add(exportContextCSV.formatCsvHeader(SInfoResultType.riverside));
        header.add(exportContextCSV.formatCsvHeader(SInfoResultType.floodDuration));
        header.add(exportContextCSV.msgUnitCSV(SInfoResultType.floodDischarge, SInfoResultType.floodDischarge.getUnit()));
        header.add(exportContextCSV.msgUnitCSV(SInfoResultType.infrastructureHeight, river.getWstUnit()));
        header.add(exportContextCSV.formatCsvHeader(SInfoResultType.infrastructuregroup));
        header.add(exportContextCSV.formatCsvHeader(SInfoResultType.infrastructuretype));

        // add dynamic headers
        final int waterlevelCount = // results.
                getWaterlevelCount();
        for (int i = 0; i < waterlevelCount; i++) {
            final String appendIndex = "_" + Integer.toString(i + 1);
            header.add(exportContextCSV.msg(DurationWaterlevel.getHeaderWCsv(), appendIndex, river.getWstUnit()));
            header.add(exportContextCSV.msg(DurationWaterlevel.getHeaderFloodDurPerYearCsv(), appendIndex));
            header.add(exportContextCSV.msg(DurationWaterlevel.getHeaderQ(), appendIndex));
            header.add(exportContextCSV.msg(DurationWaterlevel.getHeaderBezeichnCsv(), appendIndex));
        }

        if (isUseWspl())
            header.add(exportContextCSV.formatCsvHeader(GeneralResultType.gaugeLabel));

        header.add(exportContextCSV.formatCsvHeader(GeneralResultType.location));

        exportContextCSV.writeCSVLine(header.toArray(new String[header.size()]));
    }

    public List<Double> getValidDurationChartKilometers() {

        // Wunsch BfG (Testung MS3): nur stationen mit infrastruktur
        final Predicate filter = new HasInfrastructurePredicate();
        return getValues(GeneralResultType.station, filter);
    }

    @Override
    protected String[] formatCSVRow(final ExportContextCSV exportContextCSV, final ResultRow row) {

        return this.formatRow(exportContextCSV, row, ExportMode.csv);
    }

    @Override
    protected String[] formatPDFRow(final ExportContextPDF exportContextPDF, final ResultRow row) {

        return this.formatRow(exportContextPDF, row, ExportMode.pdf);
    }

    @Override
    protected void addJRTableHeader(final ExportContextPDF exportContextPDF, final MetaAndTableJRDataSource source) {

        /* column headings */
        exportContextPDF.addJRMetadata(source, "station_header", GeneralResultType.station);
        exportContextPDF.addJRMetadata(source, "riverside_header", SInfoResultType.riverside);
        exportContextPDF.addJRMetadata(source, "inundationduration_header", SInfoResultType.floodDuration);
        exportContextPDF.addJRMetadata(source, "inundationduration_q_header", SInfoResultType.floodDischarge);
        exportContextPDF.addJRMetadata(source, "infrastructure_height_header", SInfoResultType.infrastructureHeight);
        // FIXME Tironi: report-keys ..type => ..group und ..part => ..type umbenennen zwecks Einheitlichkeit
        exportContextPDF.addJRMetadata(source, "infrastructure_type_header", SInfoResultType.infrastructuregroup);
        exportContextPDF.addJRMetadata(source, "infrastructure_part_header", SInfoResultType.infrastructuretype);

        for (int i = 1; i <= this.getWaterlevelCount(); i++) {

            final String appendIndex = "_" + Integer.toString(i);
            exportContextPDF.addJRMetadata(source, getPdfHeader("w", i), exportContextPDF.msg(DurationWaterlevel.getHeaderWPdf(), appendIndex));
            exportContextPDF.addJRMetadata(source, getPdfHeader("duration", i),
                    exportContextPDF.msg(DurationWaterlevel.getHeaderFloodDurPerYearPdf(), appendIndex));
            exportContextPDF.addJRMetadata(source, getPdfHeader("q", i), exportContextPDF.msg(DurationWaterlevel.getHeaderQ(), appendIndex));
            exportContextPDF.addJRMetadata(source, getPdfHeader("bezeichnung", i),
                    exportContextPDF.msg(DurationWaterlevel.getHeaderBezeichnPdf(), appendIndex));
        }

        exportContextPDF.addJRMetadata(source, "gauge_header", GeneralResultType.gaugeLabel);
        exportContextPDF.addJRMetadata(source, "location_header", GeneralResultType.location);
    }

    private final String getPdfHeader(final String rootStr, final int index) {
        final String hd = "_header";
        final StringBuilder builder = new StringBuilder();
        return builder.append(rootStr).append("_").append(index).append(hd).toString();
    }

    public final int getWaterlevelCount() {
        return (this.waterlevelLabels != null) ? this.waterlevelLabels.length : 0;
    }

    public String getWaterlevelLabel(final int j) {
        if (this.waterlevelLabels != null && j < this.waterlevelLabels.length)
            return this.waterlevelLabels[j];
        return "";
    }

    /**
     * Gets the longitudinal section of a result value type for one river side
     */
    public final double[][] getInfrastructurePoints(final IResultType type, final Infrastructure infrastructure) {

        return getPoints(GeneralResultType.station, type, new InfrastructurePredicate(infrastructure));
    }

    /**
     * Gets a longitudinal section of W, Q, or flood duration of one of the waterlevels
     */
    public final double[][] getMainValueDurationPoints(final ValueGetter valuegetter, final int dataIndex) {

        final Collection<ResultRow> rows = getAllRows();

        final TDoubleArrayList xPoints = new TDoubleArrayList(rows.size());
        final TDoubleArrayList yPoints = new TDoubleArrayList(rows.size());

        for (final ResultRow row : rows) {

            final double station = row.getDoubleValue(GeneralResultType.station);

            final List<DurationWaterlevel> waterlevels = (List<DurationWaterlevel>) row.getValue(SInfoResultType.customMultiRowColWaterlevel);
            final DurationWaterlevel waterlevel = waterlevels.get(dataIndex);

            final Double value = valuegetter.getValue(waterlevel);

            xPoints.add(station);
            yPoints.add(value);
        }

        return new double[][] { xPoints.toNativeArray(), yPoints.toNativeArray() };
    }

    public boolean isUseWspl() {
        return this.isUseWspl;
    }

    public Collection<Entry<String, String>> getUniqueInfrastruktureTypes() {
        // TODO Auto-generated method stub
        return null;
    }
}